#include <exec/tasks.h>
#include <exec/semaphores.h>
#include <exec/execbase.h>
#include <exec/exec_protos.h>

#define SS_EXCLUSIVE 1
#define SS_SHARED    0

/* Private structure used by ObtainSemaphore() - allocated from stack */
struct SemaphoreRequest
{
    struct MinNode sr_Node;
    struct Task    *sr_Waiter;
    WORD           sr_Type;
};

/*
sr_Node			- Node for wait-linking
sr_Waiter		- Filled by the waiter
- NULLed when waked to obtain lock (so, false SIGF_DELAYs don't hurt)

ss_Node.ln_Next		- Exec linking
ss_Node.ln_Prev		- Exec linking
ss_Node.ln_Type		- NT_SEMAPHORE
ss_Node.ln_Pri		- Priority for public Semaphore
ss_Node.ln_Name		- Semaphore name for public Semaphore
ss_NestCount		- Number of locks on this semaphore
ss_WaitQueue		- List for waiters
ss_Owner		- The current owner of the semaphore
                          or NULL for shared lock
ss_QueueCount		- Number of tasks in the queue
*/

/*#define SysBase (*((struct ExecBase **)4))*/
extern struct ExecBase *SysBase;

/*
   Adds a semaphore to the system's Public Semaphores list.
 */
void _AddSemaphore(struct SignalSemaphore *s,const char *name,BYTE pri)
{
    InitSemaphore(s);
    s->ss_Node.ln_Name = (char *)name;
    s->ss_Node.ln_Pri = pri;
    Forbid();
    Enqueue(&SysBase->SemaphoreList, &s->ss_Node);
    Permit();
}

/*
   Tries to get an exclusive lock on the semaphore without blocking.
 */
LONG _AttemptSemaphore(struct SignalSemaphore *s)
{
    Forbid();
    /* Is this locked already ? */
    if(s->ss_NestCount)
    {
	Permit();
	return 0;
    }
    else
    {
	s->ss_NestCount++;
	s->ss_Owner = FindTask(NULL);
	Permit();
	return -1;
    }
}

/*
   Tries to get a share lock on the semaphore without blocking.
 */
LONG _AttemptSemaphoreShared(struct SignalSemaphore *s)
{
    Forbid();
    /* Is this exlusively locked semaphore ? */
    if(s->ss_Owner)
    {
	Permit();
	return 0;
    }
    else
    {
	s->ss_NestCount++;
	Permit();
	return -1;
    }
}

/*
   Searches the system's Public Semaphores list for a semaphore.
 */
struct SignalSemaphore *_FindSemaphore(const char *name)
{
    struct SignalSemaphore *sem = (struct SignalSemaphore *)
	FindName(&SysBase->SemaphoreList, (UBYTE *)name);

    if(sem)
    {
	sem->ss_Public++;
	/* I hope this will evaluate to an atomic instruction.
	   At least it does in gcc: addqw #1,a0@(14) */
    }
    return sem;
}

void _FreeSemaphore(struct SignalSemaphore *sem)
{
    if(sem)
    {
	sem->ss_Public--;
	/* I hope this will evaluate to an atomic instruction.
	   At least it does in gcc: subqw #1,a0@(14) */
    }
}


/*
   Initializes the semaphore structure.
 */
void _InitSemaphore(struct SignalSemaphore *s)
{
    s->ss_Node.ln_Type = NT_SEMAPHORE;
    s->ss_NestCount = 0;
    s->ss_QueueCount = 0;
    NewList(&s->ss_WaitQueue);
    s->ss_Owner = NULL;
}


/*
   Gets an exclusive (write) lock on the semaphore.
*/
void _ObtainSemaphore(struct SignalSemaphore *s)
{
    Forbid();
    /* Special case: If the semaphore is already exclusively locked by us */
    if(s->ss_Owner == FindTask(NULL))
    {
	s->ss_NestCount++;
	Permit();
	return;
    }
    /* One (or several) other lockers already ? (shared or exclusive) */
    /* If the semaphore is not available immediately, emit a request. */
    while(s->ss_NestCount)
    {
	struct SemaphoreRequest sr;

	sr.sr_Waiter = FindTask(NULL);
	sr.sr_Type   = SS_EXCLUSIVE;
	/*
	   First time around, maintain the queueing order.
	   This allows other exclusive and shared lockers to get the
	   lock before us and thus not starve to death.

	   The next time should not come, because when we are wakened,
	   there are no shared lockers or exclusive lockers on the
	   semaphore and only us are wakened.

	   Well, if while() is entered twice, we are f*cked anyway,
	   so why not add us to the tail and wait some more.
	*/
	AddTail(&s->ss_WaitQueue, (struct Node *)&sr.sr_Node);
	s->ss_QueueCount++;

	while(sr.sr_Waiter)
	{
	    Wait(SIGF_DELAY); /* Wait() breaks Forbid() */
	}
    }
    s->ss_NestCount++;
    s->ss_Owner = FindTask(NULL);
    Permit();
}


/*
   Gets a shared (read) lock on the semaphore.
*/
void _ObtainSemaphoreShared(struct SignalSemaphore *s)
{
    Forbid();
    while(s->ss_Owner) /* Exclusive onwer(s) */
    {
	struct SemaphoreRequest sr;

	sr.sr_Waiter = FindTask(NULL);
	sr.sr_Type   = SS_SHARED;
	/*
	   First time around, maintain the queueing order.
	   This allows other exclusive and shared lockers to get the
	   lock before us and thus not starve to death.

	   The next time should not come, because when we are wakened,
	   there are no exclusive lockers on the semaphore and we can
	   obtain a shared lock with anyone that happens to have or
	   wants to have a shared lock with us.

	   Well, if while() is entered twice, we are f*cked anyway,
	   so why not add us to the tail and wait some more.
	*/
	AddTail(&s->ss_WaitQueue, (struct Node *)&sr.sr_Node);
	s->ss_QueueCount++;

	while(sr.sr_Waiter)
	{
	    Wait(SIGF_DELAY); /* Wait() breaks Forbid() */
	}
    }
    s->ss_NestCount++;
    Permit();
}

/*
   Releases an obtained lock on the semaphore, whether shared or exclusive.
 */
void _ReleaseSemaphore(struct SignalSemaphore *s)
{
    Forbid();
    /*
       If the lock was shared, all shared lockers can obtain the lock.
       No need to wake up anyone if there are tasks holding a shared
       lock after we are done with it.

       However, if we are the last one holding a lock on the semaphore,
       we have to wake up the waiting exclusive locker or all of the
       waiting shared lockers.
     */
    if(--s->ss_NestCount == 0)
    {
	s->ss_Owner = NULL; /* This is valid for both shared and excl. locks */
	if(s->ss_QueueCount) /* Wake up the next locker(s) */
	{
	    struct SemaphoreRequest *sr;

	    if((sr = (struct SemaphoreRequest *)RemHead(&s->ss_WaitQueue)))
	    {
		/* Wake up the first waiter, whatever the type */
		Signal(sr->sr_Waiter, SIGF_DELAY);
		/*
		   The task is awakened here, but because we are in
		   Forbid(), no switch is done yet.
		 */
		sr->sr_Waiter = NULL;
		s->ss_QueueCount--;

		if(sr->sr_Type == SS_SHARED)
		{
		    /* If we had a shared locker, wake up all other
		       subsequent shared lockers */
		    while((sr = (struct SemaphoreRequest *)RemHead(&s->ss_WaitQueue)))
		    {
			if(sr->sr_Type == SS_SHARED)
			{
			    Signal(sr->sr_Waiter, SIGF_DELAY);
			    /*
			       The task is awakened here, but because we are in
			       Forbid(), no switch is done
			    */
			    sr->sr_Waiter = NULL;
			    s->ss_QueueCount--;
			}
			else
			{
			    /* add back the exclusive locker */
			    AddHead(&s->ss_WaitQueue,
				    (struct Node *)&sr->sr_Node);
			    break;
			}
		    }
		}
	    }
	    else
	    {
		/* strange..! */
		s->ss_QueueCount = 0;
	    }
	}
    }
    Permit();
}


/*
   Removes a semaphore from the system's Public Semaphores list.
   OBSERVE: there may still be locks held on the semaphore and
            maybe also requests pending.
 */
void _RemSemaphore(struct SignalSemaphore *s)
{
    Forbid();
    Remove(&s->ss_Node);
    /* Only removes from the Exec public list */
    Permit();
}


