#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/tasks.h>
#include "serial.h"
#include <exec/execbase.h>
#include <exec/exec_protos.h>

/* These are called from the device library-base */
LONG SerialOpen(ULONG unit, struct StdIOReq *req, ULONG flags);
void SerialClose(struct StdIOReq *req);
void SerialBeginIO(struct StdIOReq *req);
void SerialAbortIO(struct StdIOReq *req);

/*-------------------------------------------*/
static struct Unit *OpenSerial(UWORD channel,struct StdIOReq *req);
static void	CloseSerial(struct Unit *ser);
/*-------------------------------------------*/

/* Open() io_Data = serial speed */
/* in CMD_SETPARAMS io_Length = serial flags, io_Data = serial speed */

static const UWORD Prescalers[8]={0x81f,0x81e,0x40e,0x206,0x102, 0x80, 0x3e,0};
static const UWORD Bauds[8]=     {  300, 1200, 2400, 4800, 9600,19200,38400,0};

#define RECEIVE_BUFFER      8192
#define RECEIVE_HYSTERESIS  256

struct Unit
{
    struct MsgPort un_Port;
    ULONG un_Flags;
    UWORD un_OpenCnt;
    struct Task *un_Owner;

    struct Interrupt Int;
    struct SCC      *SCC;
    struct SCCConf  *SCCConf;
    UWORD            Channel;
    UWORD            ILevel; /* Interrupt vector number-64 */
    UWORD            IMask;  /* Interrupt bit mask in IMR */
    UBYTE            RxBn;   /* Current Receive Buffer Number (not used) */
    UBYTE            TxBn;   /* Current Transmit Buffer Number */

    struct List ReadRequests;
    struct List WriteRequests;

    struct StdIOReq *writereq;
    ULONG writtensize;
    struct StdIOReq *readreq;

    UBYTE *internal_buffer;  /* Location of the on-chip buf, 0x650000(+256) */
    UBYTE *receive_buffer;   /* DRAM receive buffer */
    UWORD receive_size;      /* DRAM buffer size */
    UWORD receive_top;
    UWORD receive_bot;
    UWORD received;          /* used space in receive buffer */

    UWORD lines;
    UWORD status;
};

struct SerialDevice
{
    struct Device sd_Device;
    struct Unit *sd_Units[2]; /* Pointers to units, or NULL */
};


/* Returns -1 for buffer overrun */
int  AddToBuffer(struct Unit *un, char *data, int len)
{
    int tmp;

    if(un->receive_top >= un->receive_bot)
    {
	/* |  b***t  | */
	tmp = un->receive_size - un->receive_top; /* Size left at top */
	if(len<tmp)
	{
	    memcpy(un->receive_buffer+un->receive_top, data, len);
	    un->received += len;
	    un->receive_top += len;
	    /* |  b*****t| */
	    return 0;
	}

	memcpy(un->receive_buffer+un->receive_top, data, tmp);
	un->received += tmp;
	un->receive_top = 0;
	/* |t    b***| */
	len -= tmp;
	data += tmp;
	if(!len)
	    return 0;
	/* Fall through !! */
    }

    /* |**t  b***| */
    tmp = un->receive_bot - un->receive_top; /* Size left at middle */

    if(len>tmp)
    {
	memcpy(un->receive_buffer+un->receive_top, data, tmp);
	un->received += tmp;
	un->receive_top += tmp;
	return -1;
    }
    memcpy(un->receive_buffer+un->receive_top, data, len);
    un->received += len;
    un->receive_top += len;
    return 0;
}

int  GetFromBuffer(struct Unit *un, char *data, int len)
{
    int tmp;

    if(un->receive_top < un->receive_bot)
    {
	/* |***t  b**| */
	tmp = un->receive_size - un->receive_bot;

	if(tmp > len)
	{
	    memcpy(data, un->receive_buffer+un->receive_bot, len);
	    un->received -= len;
	    un->receive_bot += len;
	    /* |***t   b*| */
	    return 0;
	}
	memcpy(data, un->receive_buffer+un->receive_bot, tmp);
	un->received -= tmp;
	un->receive_bot = 0;
	len -= tmp;
	data += tmp;
	if(!len)
	    return 0;
	/* Fall through !! */
    }
    /* |b**t    | */
    tmp = un->receive_top - un->receive_bot;

    if(tmp > len)
    {
	memcpy(data, un->receive_buffer+un->receive_bot, len);
	un->received -= len;
	un->receive_bot += len;
	/* | b*t    | */
	return 0;
    }
    /* Buffer becomes empty */
    /* We assume that the user won't try to read more than we have.. */
    /* If the user DOES try to read more, he will still only get what
       we have and nothing more. */
    memcpy(data, un->receive_buffer+un->receive_bot, tmp);
    un->received = 0;
    un->receive_bot = un->receive_top = 0;
    return 0;
}


LONG SerialOpen(ULONG unit, struct StdIOReq *req, ULONG flags)
{
    struct Unit *temp;

/* TODO: locking */

    /* Is the unit currently in use ? */
    temp = ((struct SerialDevice *)(req->io_Device))->sd_Units[unit];
    if(temp)
    {
	/* Check if we can grant Shared access */
	if( ((temp->un_Flags & SERF_EXCLUSIVE) &&
	     temp->un_Owner != FindTask(NULL) ) ||
	   (flags & SERF_EXCLUSIVE))
	{
	    req->io_Error = 1;
	}
	else
	{
	    req->io_Unit =
		((struct SerialDevice *)req->io_Device)->sd_Units[unit];
	    req->io_Unit->un_OpenCnt++;
	    req->io_Error = 0;
	    req->io_Device->dd_Library.lib_OpenCnt++;
	}
    }
    else
    {
	if(temp = OpenSerial(unit,req))
	{
	    temp->un_Owner = FindTask(NULL);
	    ((struct SerialDevice *)req->io_Device)->sd_Units[unit] = temp;
	    req->io_Unit = temp;
	    req->io_Unit->un_Flags = flags;
	    req->io_Unit->un_OpenCnt++;
	    req->io_Error = 0;
	    req->io_Device->dd_Library.lib_OpenCnt++;
	}
	else
	{
	    req->io_Error = 2;
	}
    }
    return req->io_Error;
}

void SerialClose(struct StdIOReq *req)
{
    if(req->io_Unit)
    {
	req->io_Device->dd_Library.lib_OpenCnt--;
	req->io_Unit->un_OpenCnt--;
	if(!req->io_Unit->un_OpenCnt)
	{
	    CloseSerial(req->io_Unit);
	}
	req->io_Unit = NULL;
    }
}

void SerialBeginIO(struct StdIOReq *req)
{
    PutMsg(&req->io_Unit->un_Port, &req->io_Message);
}

void SerialAbortIO(struct StdIOReq *req)
{
    struct Unit *ser = (struct Unit *)req->io_Unit;

    Disable();
    if(req->io_Message.mn_Node.ln_Type == NT_REPLYMSG)
    {
	Enable();
	return;
    }
    if(req == ser->readreq)
    {
	req->io_Actual = 0;
	ReplyMsg( &req->io_Message);
	ser->readreq = (struct StdIOReq *)RemHead(&ser->ReadRequests);
	Enable();
	return;
    }
    if(req == ser->writereq)
    {
	/* Well, let the user just wait until the block is transmitted */
#if 0
	/* TODO: stop the transmitter */
	req->io_Actual = 0;
	ReplyMsg( &req->io_Message);
	if(( ser->writereq = (struct StdIOReq *)RemHead(&ser->WriteRequests)))
	{
	    ser->SCC->TxBD[0].Address = ser->writereq->io_Data;
	    if(ser->writereq->io_Length > 65535)
	    {
		ser->SCC->TxBD[0].Length = 65535;
		ser->writtensize = 65535;
	    }
	    else
	    {
		ser->SCC->TxBD[0].Length = ser->writereq->io_Length;
		ser->writtensize = ser->writereq->io_Length;
	    }
	    ser->SCC->TxBD[0].Status = 0xf000;
	    /* Do not use TxBD:s 6 and 7 in SCCS2 -
	       they are used for RAM refresh */
	}
#endif
	Enable();
	return;
    }
    Remove((struct Node *)req);
    req->io_Actual = 0;
    ReplyMsg(&req->io_Message);
    Enable();
    return;
}


void SerialHandler(ULONG is_Data, UWORD *StackFrame);
void SerialHandler(ULONG is_Data, UWORD *StackFrame)
{
    struct Unit *ser=(struct Unit *)is_Data;
    UBYTE stat = ser->SCCConf->SCCE;
    struct StdIOReq *message;
    int size,i;

    ser->SCCConf->SCCE = stat;
    if(stat & SER_RX)
    {
	ser->status &= ~SER_RX;
	i = ser->SCC->RxBD[0].Length;
	/* just to be certain */
	if(ser->received + i < ser->receive_size)
	{
	    AddToBuffer(ser, (char *)ser->internal_buffer, i);
	}
#if 0
	/* Check from the chunk we just received */
	if(ser->internal_buffer[i-1]=='\r' ||
	   ser->internal_buffer[i-1]=='\n')
	    ser->lines++;
#endif
	/* Check if the buffer is empty enough */
	if(ser->received + ser->SCC->Param.MRBLR <= ser->receive_size)
	{
	    /* Continue receiving to On-Chip RAM */
	    ser->SCC->RxBD[0].Status = 0xb000;
	    ser->status |= SER_RX;
	}
	/*
	   Otherwise we stop receiving and start receiving again when the
	   buffer gets a little less full
	 */
	while( ser->readreq &&
	       (ser->readreq->io_Length <= ser->received))
	{
	    size = ser->readreq->io_Length;

	    GetFromBuffer(ser, (char *)ser->readreq->io_Data, size);
	    ser->readreq->io_Actual = size;
	    ser->readreq->io_Error = 0;
	    ReplyMsg( &ser->readreq->io_Message );

	    ser->readreq = (struct StdIOReq *)RemHead(&ser->ReadRequests);
	}
	if( !(ser->status & SER_RX) &&
	   (ser->received + ser->SCC->Param.MRBLR <= ser->receive_size))
	{
	    /* Continue receiving to On-Chip RAM */
	    ser->SCC->RxBD[0].Status = 0xb000;
	    ser->status |= SER_RX;
	}
	*(volatile UWORD *)0x650818 = ser->IMask;  /* ISR */
	*(volatile UWORD *)0x650814 = ser->IMask;  /* IPR */
    }

    if((stat & SER_TX) && !(ser->SCC->TxBD[0].Status & 0x8000))
    {
	/* Buffer transmitted */
	if( ser->writereq )
	{
	    size = ser->writereq->io_Length - ser->writtensize;
	    if(size > 0)
	    {
		ser->SCC->TxBD[0].Address =
		    ser->writereq->io_Data + ser->writtensize;
		if(size > 65535)
		    size = 65535; /* 0 would be exactly 64k */
		ser->SCC->TxBD[0].Length = size;
		ser->SCC->TxBD[0].Status = 0xf000;
		/* Do not use TxBD:s 6 and 7 in SCCS2 -
		   they are used for RAM refresh */
		ser->writtensize += size;
	    }
	    else
	    {
		ser->writereq->io_Error = 0;
		ser->writereq->io_Actual = ser->writtensize;
		ReplyMsg( &ser->writereq->io_Message );
		ser->writereq = NULL;
	    }
	}
	if(!ser->writereq && 
	   (message = (struct StdIOReq *)RemHead(&ser->WriteRequests)))
	{
	    ser->writereq = message;
	    ser->SCC->TxBD[0].Address = message->io_Data;
	    size = message->io_Length;
	    if(size > 65535)
		size = 65535; /* 0 would be exactly 64k */
	    ser->SCC->TxBD[0].Length = size;
	    ser->SCC->TxBD[0].Status = 0xf000;
	    /* Do not use TxBD:s 6 and 7 in SCCS2 -
	       they are used for RAM refresh */
	    ser->writtensize = size;
	}
	/* Else someone else sent a buffer !? */
	*(volatile UWORD *)0x650818 = ser->IMask;  /* ISR */
	*(volatile UWORD *)0x650814 = ser->IMask;  /* IPR */
    }

    if(stat & SER_BRK)
    {
	ser->status |= SER_BRK;
	*(volatile UWORD *)0x650818 = ser->IMask;  /* ISR */
	*(volatile UWORD *)0x650814 = ser->IMask;  /* IPR */
    }

    while( (message = (struct StdIOReq *)GetMsg(&ser->un_Port)) )
    {
	message->io_Error = 10;
	switch(message->io_Command)
	{
	case CMD_READ:
	    size = message->io_Length;
	    if(!size)
	    {
		message->io_Actual = 0;
		message->io_Error = 0;
		ReplyMsg( &message->io_Message );
		break;
	    }
	    if( size <= ser->received )
	    {
		GetFromBuffer(ser, (char *)message->io_Data, size);
		message->io_Actual = size;
		message->io_Error = 0;
		ReplyMsg( &message->io_Message );

		/* Check if the buffer is empty enough */
		if( !(ser->status & SER_RX) &&
		   (ser->received + ser->SCC->Param.MRBLR <= ser->receive_size))
		{
		    /* Continue receiving to On-Chip RAM */
		    ser->SCC->RxBD[0].Status = 0xb000;
		    ser->status |= SER_RX;
		}
	    }
	    else
	    {
		if( ser->readreq )
		{
		    /* Someone already waiting for data */
		    AddTail( &ser->ReadRequests,&message->io_Message.mn_Node );
		}
		else
		{
		    ser->readreq = message;
		}
	    }
	    break;
	case CMD_WRITE:
	    if(!message->io_Length)
	    {
		message->io_Actual = 0;
		message->io_Error = 0;
		ReplyMsg( &message->io_Message );
		break;
	    }
	    if( ser->writereq || (ser->SCC->TxBD[0].Status & 0x8000))
	    {
		/* transmitter/buffer busy */
		AddTail( &ser->WriteRequests, &message->io_Message.mn_Node );
	    }
	    else
	    {
		/* transmitter ready */
		ser->writereq = message;
		ser->SCC->TxBD[0].Address = message->io_Data;
		size = message->io_Length;
		if(size > 65535)
		    size = 65535; /* 0 would be exactly 64k */
		ser->SCC->TxBD[0].Length = size;
		ser->SCC->TxBD[0].Status = 0xf000;
		/* Do not use TxBD:s 6 and 7 in SCCS2 -
		   they are used for RAM refresh */
		ser->writtensize = size;
	    }
	    break;
	case SDCMD_BREAK:
	    /* TODO without busy looping */
	    ser->SCC->Param.BRKCR = message->io_Data;
	    while( CR & 1 );
	    CR = ser->Channel*2 + 0x01; /* STOP TRANSMIT */
	    while( CR & 1 );
	    CR = ser->Channel*2 + 0x11; /* RESTART TRANSMIT */
	    ser->SCC->Param.BRKCR = 0;
	    message->io_Error = 0;
	    ReplyMsg( &message->io_Message );
	    break;
	case CMD_SETPARAMS:
#if 0
	    /* Changing the speed means we have to stop receiver and
	       transmitter and enable them after we have set the new speed. */
	    i=0;
	    while(Bauds[i] && Bauds[i]<message->io_Length)
		i++;
	    if(Bauds[i])
	    {
		ser->SCCConf->SCON = Prescalers[i];
	    }
	    else
	    {
		message->io_Error = 10;
	    }
#endif
	    ser->un_Flags = message->io_Data;
	    message->io_Error = 0;
	    ReplyMsg( &message->io_Message );
	    break;
	case SDCMD_QUERY:
	    message->io_Actual = ser->received;
	    ReplyMsg( &message->io_Message );
	    break;
	case CMD_FLUSH:
	    ReplyMsg( &message->io_Message );
	    break;
	case CMD_RESET:
	    ReplyMsg( &message->io_Message );
	    break;
	default:
	    ReplyMsg( &message->io_Message );
	    break;
	}
    }
    *(volatile UBYTE *)0xf4 = 8;  /* Clear IPA in SCR */
}

static struct Unit *OpenSerial(UWORD channel,struct StdIOReq *req)
{
    struct Unit *ser = (struct Unit *)AllocMem(sizeof(struct Unit)+RECEIVE_BUFFER, 0);
    int i;

    if(!ser)
	return NULL;

    ser->receive_buffer = (UBYTE *)ser+sizeof(struct Unit);
    ser->receive_size = RECEIVE_BUFFER;
    ser->received = 0;
    ser->receive_bot = ser->receive_top = 0;
    ser->lines = 0;
    ser->readreq = NULL;
    ser->writereq = NULL;
    NewList( &ser->ReadRequests );
    NewList( &ser->WriteRequests );
    ser->Channel = channel;
    ser->status = 0;

    if(channel == 0)
    {
	ser->SCC = (struct SCC *)0x650400;
	ser->SCCConf = (struct SCCConf *)0x650880;
	ser->ILevel = 13;
	ser->IMask = 0x2000;
    }
    else
    if(channel == 1)
    {
	ser->SCC = (struct SCC *)0x650500;
	ser->SCCConf = (struct SCCConf *)0x650890;
	ser->ILevel = 10;
	ser->IMask = 0x0400;
    }
    else
    {
	FreeMem(ser, sizeof(struct Unit)+ser->receive_size);
	return 0;
    }

    i=0;
    while(Bauds[i] && Bauds[i]<req->io_Data)
	i++;

    if(!Bauds[i])
    {
	FreeMem(ser, sizeof(struct Unit)+ser->receive_size);
	return NULL;
    }

    Forbid();
    CR = ser->Channel*2 | 1; /* STOP TRANSMIT */
    while( CR & 1 );
    Permit();
    /* 'Manual' mode, Asynchronous, Tx&Rx Disabled, 8N1 */
    ser->SCCConf->SCM = 0x0131;
    ser->SCCConf->SCON = Prescalers[i];
    /* ser->SCCConf->DSR = 0x7e7e (Fractional stop bits if needed) */
    ser->SCC->Param.RFCR = ser->SCC->Param.TFCR = 0;
    ser->SCC->Param.MRBLR = 256;  /* Receive buffer size */
    ser->SCC->Param.MAX_IDL = 16; /* Close buffer after 16 char idle time */

#if 0
    ser->SCC->Param.CHARACTER[0] = '\n'; /* Close buffer & cause intr. */
    ser->SCC->Param.CHARACTER[1] = '\r';
#else
    ser->SCC->Param.CHARACTER[0] = 0x8000; /* No special characters */
#endif
    ser->SCC->Param.CHARACTER[2] = 0x8000;
    ser->SCC->Param.CHARACTER[8] = 0x0000;

    ser->RxBn = ser->TxBn = 0;   /* Not used */
    ser->SCC->Param.RBDnum = 0;
    ser->SCC->Param.TBDnum = 0x40;

    ser->SCC->RxBD[0].Status = 0x3000; /* Only one BD used, internal address*/
    ser->SCC->RxBD[0].Address = (ULONG)(channel*256);
    ser->internal_buffer = (UBYTE *)(0x650000+channel*256);
    ser->SCC->TxBD[0].Status = 0x7000; /* Only one BD used, external address */

    ser->Int.is_Code = &SerialHandler;
    ser->Int.is_Data = (ULONG)ser;
    AddIntServer(64+ser->ILevel, &ser->Int);

    NewList(&ser->un_Port.mp_MsgList);
    ser->un_Port.mp_Flags = PA_SOFTINT;
    ser->un_Port.mp_SoftInt = (void *)(64+ser->ILevel);  /* The same handler */

    ser->SCCConf->SCCE = 0xff;
    ser->SCCConf->SCCM = 0x03;        /* Enable Tx and Rx interrupts */
    *(volatile UWORD *)0x650818 = ser->IMask;  /* ISR */
    *(volatile UWORD *)0x650814 = ser->IMask;  /* IPR */
    *(volatile UWORD *)0x650816 |= ser->IMask; /* IMR */

    Forbid();
    CR = ser->Channel*2 | 0x11; /* Restart transmit command */
    while(CR & 1);
    CR = ser->Channel*2 | 0x21; /* Enter Hunt mode */
    while(CR & 1);
    Permit();
    ser->SCCConf->SCM = 0x0131 | 0xc; /* Enable Tx and Rx */
 
    /* Start receiving to On-Chip RAM */
    ser->SCC->RxBD[0].Status = 0xb000;
    ser->status |= SER_RX;

    return ser;
}

static void CloseSerial(struct Unit *ser)
{
    *(volatile UWORD *)0x650816 &= ~ser->IMask; /* IMR */
    RemIntServer(64+ser->ILevel, &ser->Int);
    FreeMem(ser, sizeof(struct Unit)+ser->receive_size);
}

void AddSerialDevice()
{
    ULONG memchunk = (ULONG)AllocMem(sizeof(struct SerialDevice)+24, 0);
    /* allocate struct device and lib_negsize */
    struct SerialDevice *serdev;
    static const ULONG vectors[4] = {
	(ULONG)SerialOpen, (ULONG)SerialClose,
	(ULONG)SerialBeginIO, (ULONG)SerialAbortIO
    };

    if(memchunk)
    {
	serdev = (struct SerialDevice *)(memchunk+24);

	InitLibFuncs(&serdev->sd_Device.dd_Library, (ULONG *)vectors, 4);
	serdev->sd_Device.dd_Library.lib_Node.ln_Name = "serial.device";
	serdev->sd_Device.dd_Library.lib_Version = 1235;
	serdev->sd_Device.dd_Library.lib_OpenCnt = 0;
	serdev->sd_Units[0] = NULL;
	serdev->sd_Units[1] = NULL;
	AddDevice( &serdev->sd_Device );
    }
}
